import inspect
import pprint
import re
from typing import TYPE_CHECKING

import lale.helpers

if TYPE_CHECKING:
    from lale.operators import IndividualOp


def _indent(prefix, string, first_prefix=None):
    lines = string.splitlines()
    if lines:
        if first_prefix is None:
            first_prefix = prefix
        first_indented = (first_prefix + lines[0]).rstrip()
        rest_indented = [(prefix + line).rstrip() for line in lines[1:]]
        result = first_indented + "\n" + "\n".join(rest_indented)
        return result
    else:
        return ""


def _value_docstring(value):
    return pprint.pformat(value, width=10000, compact=True)


def _kind_tag(schema):
    if "anyOf" in schema:
        return "union type"
    elif "allOf" in schema:
        return "intersection type"
    elif "not" in schema or "laleNot" in schema:
        return "negated type"
    elif "type" in schema:
        if schema["type"] == "object":
            return "dict"
        elif schema["type"] == "number":
            return "float"
        elif isinstance(schema["type"], list):
            return " *or* ".join(schema["type"])
        else:
            return schema["type"]
    elif "enum" in schema:
        values = schema["enum"]
        assert len(values) >= 1
        if len(values) == 1:
            return _value_docstring(values[0])
        elif len(values) == 2:
            return " *or* ".join([_value_docstring(v) for v in values])
        else:
            prefix = ", ".join([_value_docstring(v) for v in values[:-1]])
            suffix = ", *or* " + _value_docstring(values[-1])
            return prefix + suffix
    else:
        return "any type"


def _schema_docstring(name, schema, required=True, relevant=True):
    tags = []
    if "laleType" in schema:
        tags.append(schema["laleType"])
    else:
        tags.append(_kind_tag(schema))
    if "minimum" in schema:
        op = ">" if schema.get("exclusiveMinimum", False) else ">="
        tags.append(op + _value_docstring(schema["minimum"]))
    if "minimumForOptimizer" in schema:
        tags.append(
            ">=" + _value_docstring(schema["minimumForOptimizer"]) + " for optimizer"
        )
    if "maximum" in schema:
        op = "<" if schema.get("exclusiveMaximum", False) else "<="
        tags.append(op + _value_docstring(schema["maximum"]))
    if "laleMaximum" in schema:
        tags.append("<=" + _value_docstring(schema["laleMaximum"]))
    if "maximumForOptimizer" in schema:
        tags.append(
            "<=" + _value_docstring(schema["maximumForOptimizer"]) + " for optimizer"
        )
    if "distribution" in schema:
        tags.append(schema["distribution"] + " distribution")
    if "minItems" in schema:
        tags.append(">=" + _value_docstring(schema["minItems"]) + " items")
    if "minItemsForOptimizer" in schema:
        tags.append(
            ">="
            + _value_docstring(schema["minItemsForOptimizer"])
            + " items for optimizer"
        )
    if "maxItems" in schema:
        tags.append("<=" + _value_docstring(schema["maxItems"]) + " items")
    if "maxItemsForOptimizer" in schema:
        tags.append(
            "<="
            + _value_docstring(schema["maxItemsForOptimizer"])
            + " items for optimizer"
        )
    if not required:
        tags.append("optional")
    if not relevant or schema.get("forOptimizer", True) is False:
        tags.append("not for optimizer")
    if schema.get("transient", False):
        tags.append("transient")
    if "default" in schema:
        tags.append("default " + _value_docstring(schema["default"]))

    def item_docstring(name, item_schema, required=True):
        sd = _schema_docstring(name, item_schema, required=required)
        return _indent("    ", sd, "  - ").rstrip()

    body = None
    if "anyOf" in schema:
        item_docstrings = [item_docstring(None, s) for s in schema["anyOf"]]
        if name is not None and name.startswith("_`constraint-"):
            rexp = re.compile(r"^(  - )(dict \*of\* )(.+)")
            item_docstrings = [rexp.sub(r"\1\3", s) for s in item_docstrings]
        if len(item_docstrings) > 1:
            rexp = re.compile(r"^(  - )(.+)")
            rest = [rexp.sub(r"\1*or* \2", s) for s in item_docstrings[1:]]
            item_docstrings = item_docstrings[:1] + rest
        body = "\n\n".join(item_docstrings)
    elif "allOf" in schema:
        item_docstrings = [item_docstring(None, s) for s in schema["allOf"]]
        if len(item_docstrings) > 1:
            rexp = re.compile(r"^(  - )(.+)")
            rest = [rexp.sub(r"\1*and* \2", s) for s in item_docstrings[1:]]
            item_docstrings = item_docstrings[:1] + rest
        body = "\n\n".join(item_docstrings)
    elif "not" in schema:
        body = item_docstring(None, schema["not"])
    elif "laleNot" in schema:
        body = f"  - '{schema['laleNot']}'"
    elif schema.get("type", "") == "array":
        if "items" in schema:
            items_schemas = schema["items"]
            if isinstance(items_schemas, dict):
                body = item_docstring("items", items_schemas)
            else:
                items_docstrings = [
                    item_docstring(f"item {i}", s) for i, s in enumerate(items_schemas)
                ]
                body = "\n\n".join(items_docstrings)
    elif schema.get("type", "") == "object" and "properties" in schema:
        item_docstrings = [
            item_docstring(k, s) for k, s in schema["properties"].items()
        ]
        body = "\n\n".join(item_docstrings)
    result = name + " : " if name else ""
    try:
        result += ", ".join(tags)
    except BaseException as e:
        raise ValueError(f"Unexpected internal error for {schema}.") from e
    assert len(result) > 0 and result.rstrip() == result
    if result.startswith("-"):
        result = "\\" + result
    if body is not None and body.find("\n") == -1:
        assert body.startswith("  - ")
        result += " *of* " + body[4:]
    if "description" in schema:
        result += "\n\n" + _indent("  ", schema["description"]).rstrip()
    if body is not None and body.find("\n") != -1:
        result += "\n\n" + body
    return result.rstrip()


def _params_docstring(params_schema, hp2constraints=None):
    if params_schema is None:
        return ""
    params = params_schema.get("properties", {})
    if len(params) == 0:
        result = ""
    else:
        result = "Parameters\n----------\n"
    for param_name, param_schema in params.items():
        required = param_name in params_schema.get("required", {})
        relevant = (
            "relevantToOptimizer" not in params_schema
            or param_name in params_schema["relevantToOptimizer"]
        )
        item_docstring = _schema_docstring(param_name, param_schema, required, relevant)
        result += _indent("  ", item_docstring, "").rstrip()
        if hp2constraints is not None and param_name in hp2constraints:
            constraints = [f"`constraint-{i}`_" for i in hp2constraints[param_name]]
            result += f"\n\n  See also {', '.join(constraints)}."
        result += "\n\n"
    return result


def _arg_docstring(val):
    if val is None:
        return str("None")
    if isinstance(val, (int, float)):
        return str(val)
    elif isinstance(val, list):
        return [_arg_docstring(x) for x in val]
    elif isinstance(val, dict):
        return {_arg_docstring(k): _arg_docstring(v) for k, v in val.items()}
    else:
        return f'"{str(val)}"'


def _paramlist_docstring(hyperparams_schema) -> str:
    params = hyperparams_schema.get("allOf", None)
    if params is None:
        return ""
    if isinstance(params, list):
        if not params:
            return ""
        params = params[0]
    if params is None:
        return ""
    params = params.get("properties", {})
    if len(params) == 0:
        return ""
    result = ", *"
    for param_name, param_schema in params.items():
        result += f", {param_name}"
        default = param_schema.get("default", None)
        if "default" in param_schema:
            default = param_schema["default"]
            default_str = _arg_docstring(default)
            if default_str is not None:
                result += f"={default_str}"
    return result


def _get_hp2constraints(hyperparams_schema):
    result = {}
    for i in range(1, len(hyperparams_schema["allOf"])):
        schema = hyperparams_schema["allOf"][i]
        for disjunct in schema.get("anyOf", []):
            for hyperparam in disjunct.get("properties", {}).keys():
                result[hyperparam] = result.get(hyperparam, []) + [i]
    return result


def _hyperparams_docstring(hyperparams_schema):
    hp2constraints = _get_hp2constraints(hyperparams_schema)
    result = _params_docstring(hyperparams_schema["allOf"][0], hp2constraints)
    if len(hyperparams_schema["allOf"]) > 1:
        result += "Notes\n-----\n"
        item_docstrings = [
            _schema_docstring(f"_`constraint-{i}`", hyperparams_schema["allOf"][i])
            for i in range(1, len(hyperparams_schema["allOf"]))
        ]
        result += "\n\n".join(item_docstrings)
    return result


def _method_docstring(description, ready_string, params_schema, result_schema=None):
    result = description + "\n\n"
    if ready_string is not None:
        result += "*Note: " + ready_string + "*\n\n"
        result += (
            "Once this method is available, it will have the following signature: \n\n"
        )
    result += _params_docstring(params_schema)
    if result_schema is not None:
        result += "Returns\n-------\n"
        item_docstring = _schema_docstring("result", result_schema)
        result += _indent("  ", item_docstring, "")
        result += "\n\n"
    return result


def _cls_docstring(cls, combined_schemas):
    descr_lines = combined_schemas["description"].splitlines()
    result = descr_lines[0]
    result += "\n\nThis documentation is auto-generated from JSON schemas.\n\n"
    more_description = "\n".join(descr_lines[1:]).strip()
    if more_description != "":
        result += more_description + "\n\n"
    return result


def _set_docstrings_helper(cls, lale_op, combined_schemas):
    properties = combined_schemas.get("properties", None)

    assert cls.__doc__ is None
    impl_cls = lale_op.impl_class
    cls.__doc__ = _cls_docstring(impl_cls, combined_schemas)

    if properties is not None:
        hyperparams_schema = properties.get("hyperparams", None)
        if hyperparams_schema is not None:
            doc = _hyperparams_docstring(hyperparams_schema)
            try:
                args = _paramlist_docstring(hyperparams_schema)

                code = f"""
def __init__(self{args}):
    pass
"""
                import math

                d = {}
                exec(code, {"nan": math.nan, "inf": math.inf}, d)
                __init__ = d["__init__"]  # type: ignore
            except BaseException as e:
                import warnings

                warnings.warn(
                    f"""While trying to generate a docstring for {cls.__name__}, when trying
to create an init method with the appropriate parameter list, an exception was raised: {e}"""
                )

                def __init__(self):
                    pass

            __init__.__doc__ = doc
            cls.__init__ = __init__

    def make_fun(
        fun_name,
        fake_fun,
        description,
        ready_string,
        params_schema_key,
        result_schema_key=None,
    ):
        params_schema = None
        result_schema = None
        if properties is not None:
            if params_schema_key is not None:
                params_schema = properties.get(params_schema_key, None)
            if result_schema_key is not None:
                result_schema = properties.get(result_schema_key, None)

        if hasattr(impl_cls, fun_name):
            ready_string_to_use = None
            if not hasattr(cls, fun_name):
                ready_string_to_use = ready_string
            doc = _method_docstring(
                description, ready_string_to_use, params_schema, result_schema
            )
            setattr(cls, fun_name, fake_fun)
            fake_fun.__name__ = "fun_name"
            fake_fun.__doc__ = doc

    def fit(self, X, y=None, **fit_params):
        pass

    make_fun(
        "fit",
        fit,
        "Train the operator.",
        "The fit method is not available until this operator is trainable.",
        "input_fit",
    )

    def transform(self, X, y=None):
        pass

    make_fun(
        "transform",
        transform,
        "Transform the data.",
        "The transform method is not available until this operator is trained.",
        "input_transform",
        "output_transform",
    )

    def predict(self, X, **predict_params):
        pass

    make_fun(
        "predict",
        predict,
        "Make predictions.",
        "The predict method is not available until this operator is trained.",
        "input_predict",
        "output_predict",
    )

    def predict_proba(self, X):
        pass

    make_fun(
        "predict_proba",
        predict_proba,
        "Probability estimates for all classes.",
        "The predict_proba method is not available until this operator is trained.",
        "input_predict_proba",
        "output_predict_proba",
    )

    def decision_function(self, X):
        pass

    make_fun(
        "decision_function",
        decision_function,
        "Confidence scores for all classes.",
        "The decision_function method is not available until this operator is trained.",
        "input_decision_function",
        "output_decision_function",
    )


def set_docstrings(lale_op: "IndividualOp"):
    """
    If we are running under sphinx, this will take
    a variable whose value is a lale operator
    and change it to a value of an artificial class
    with appropriately documented methods.
    """
    try:
        if __sphinx_build__:  # type: ignore
            try:

                # impl = lale_op.impl_class
                frm = inspect.stack()[1]
                module = inspect.getmodule(frm[0])
                assert module is not None
                combined_schemas = lale_op._schemas
                name = lale.helpers.arg_name(pos=0, level=1)
                assert name is not None

                # we want to make sure that the Operator constructor args are not shown
                def __init__():
                    pass

                new_class = type(name, (lale_op.__class__,), {"__init__": __init__})  # type: ignore
                new_class.__module__ = module.__name__
                module.__dict__[name] = new_class

                _set_docstrings_helper(new_class, lale_op, combined_schemas)
            except NameError as e:
                raise ValueError(e)
    except NameError:
        pass



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
from sklearn.decomposition import NMF as SKLModel

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "description": "Non-Negative Matrix Factorization (NMF)",
    "allOf": [
        {
            "type": "object",
            "required": [
                "n_components",
                "init",
                "solver",
                "beta_loss",
                "tol",
                "max_iter",
                "random_state",
                "alpha",
                "l1_ratio",
                "verbose",
                "shuffle",
            ],
            "relevantToOptimizer": [
                "n_components",
                "tol",
                "max_iter",
                "alpha",
                "shuffle",
            ],
            "additionalProperties": False,
            "properties": {
                "n_components": {
                    "anyOf": [
                        {
                            "type": "integer",
                            "minimum": 1,
                            "laleMaximum": "X/items/maxItems",  # number of columns
                            "minimumForOptimizer": 2,
                            "maximumForOptimizer": 256,
                            "distribution": "uniform",
                        },
                        {
                            "description": "If not set, keep all components.",
                            "enum": [None],
                        },
                    ],
                    "default": None,
                    "description": "Number of components.",
                },
                "init": {
                    "enum": ["custom", "nndsvd", "nndsvda", "nndsvdar", "random", None],
                    "default": None,
                    "description": "Method used to initialize the procedure.",
                },
                "solver": {
                    "enum": ["cd", "mu"],
                    "default": "cd",
                    "description": "Numerical solver to use:",
                },
                "beta_loss": {
                    "description": "Beta divergence to be minimized, measuring the distance between X and the dot product WH.",
                    "anyOf": [
                        {
                            "type": "number",
                            "minimumForOptimizer": -1,
                            "maximumForOptimizer": 1,
                        },
                        {"enum": ["frobenius", "kullback-leibler", "itakura-saito"]},
                    ],
                    "default": "frobenius",
                },
                "tol": {
                    "type": "number",
                    "minimum": 0.0,
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 0.01,
                    "distribution": "loguniform",
                    "default": 0.0001,
                    "description": "Tolerance of the stopping condition.",
                },
                "max_iter": {
                    "type": "integer",
                    "minimum": 1,
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 1000,
                    "distribution": "uniform",
                    "default": 200,
                    "description": "Maximum number of iterations before timing out.",
                },
                "random_state": {
                    "anyOf": [
                        {"type": "integer"},
                        {"laleType": "numpy.random.RandomState"},
                        {"enum": [None]},
                    ],
                    "default": None,
                    "description": "Used for initialization and in coordinate descent.",
                },
                "alpha": {
                    "type": "number",
                    "minimumForOptimizer": 1e-10,
                    "maximumForOptimizer": 1.0,
                    "distribution": "loguniform",
                    "default": 0.0,
                    "description": "Constant that multiplies the regularization terms. Set it to zero to have no regularization.",
                },
                "l1_ratio": {
                    "type": "number",
                    "default": 0.0,
                    "minimum": 0.0,
                    "maximum": 1.0,
                    "description": "The regularization mixing parameter.",
                },
                "verbose": {
                    "anyOf": [{"type": "boolean"}, {"type": "integer"}],
                    "default": 0,
                    "description": "Whether to be verbose.",
                },
                "shuffle": {
                    "type": "boolean",
                    "default": False,
                    "description": "If true, randomize the order of coordinates in the CD solver.",
                },
            },
        },
        {
            "description": "beta_loss, only in 'mu' solver",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {
                        "beta_loss": {"enum": ["frobenius"]},
                    },
                },
                {
                    "type": "object",
                    "properties": {
                        "solver": {"enum": ["mu"]},
                    },
                },
            ],
        },
    ],
}

_input_fit_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"type": "number", "minimum": 0.0},
            },
        },
        "y": {"laleType": "Any"},
    },
}

_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number", "minimum": 0.0}},
        }
    },
}

_output_transform_schema = {
    "description": "Transformed data",
    "type": "array",
    "items": {
        "type": "array",
        "items": {"type": "number"},
    },
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Non-negative matrix factorization`_ transformer from scikit-learn for linear dimensionality reduction.

.. _`Non-negative matrix factorization`: https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.NMF.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.nmf.html",
    "import_from": "sklearn.decomposition",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}

NMF: lale.operators.PlannedIndividualOp
NMF = lale.operators.make_operator(SKLModel, _combined_schemas)

if sklearn.__version__ >= "0.24":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.decomposition.NMF.html
    # new: https://scikit-learn.org/0.24/modules/generated/sklearn.decomposition.NMF.html
    from lale.schemas import AnyOf, Enum, Null

    NMF = NMF.customize_schema(
        regularization=AnyOf(
            desc="Select whether the regularization affects the components (H), the transformation (W), both or none of them.",
            types=[
                Enum(values=["both", "components", "transformation"]),
                Null(),
            ],
            default="both",
            forOptimizer=True,
        ),
        set_as_available=True,
    )

if sklearn.__version__ >= "1.0":
    # old: https://scikit-learn.org/0.24/modules/generated/sklearn.decomposition.NMF.html
    # new: https://scikit-learn.org/1.0/modules/generated/sklearn.decomposition.NMF.html
    from lale.schemas import AnyOf, Enum, Float, Null

    NMF = NMF.customize_schema(
        alpha=Float(
            desc="""Constant that multiplies the regularization terms.
Set it to zero to have no regularization. When using alpha instead of alpha_W and alpha_H,
the regularization terms are not scaled by the n_features (resp. n_samples) factors for W (resp. H).""",
            default=0.0,
            forOptimizer=False,
        ),
        alpha_W=Float(
            desc="""Constant that multiplies the regularization terms of W. Set it to zero (default) to have no regularization on W.""",
            minimumForOptimizer=1e-10,
            maximumForOptimizer=1.0,
            distribution="loguniform",
            default=0.0,
            forOptimizer=True,
        ),
        alpha_H=AnyOf(
            types=[
                Enum(values=["same"]),
                Float(
                    minimumForOptimizer=1e-10,
                    maximumForOptimizer=1.0,
                    distribution="loguniform",
                ),
            ],
            desc="""Constant that multiplies the regularization terms of H.
Set it to zero to have no regularization on H. If “same” (default), it takes the same value as alpha_W.""",
            default="same",
            forOptimizer=True,
        ),
        regularization=AnyOf(
            desc="Select whether the regularization affects the components (H), the transformation (W), both or none of them.",
            types=[
                Enum(values=["both", "components", "transformation"]),
                Null(),
            ],
            default="both",
            forOptimizer=False,
        ),
        set_as_available=True,
    )

lale.docstrings.set_docstrings(NMF)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any


class Visitor(object):
    def defaultVisit(self, node, *args, **kwargs):
        raise NotImplementedError

    def __getattr__(self, attr):
        if attr.startswith("visit"):
            return self.defaultVisit
        return self.__getattribute__(attr)

    def _visitAll(self, iterable, *args, **kwargs):
        def filter(x):
            return (x is not None) or None

        return [filter(x) and accept(x, self, *args, **kwargs) for x in iterable]


# Because of the magic way we add accept methods, mypy does not know they exist
# so this method is important for accept calls to typecheck
def accept(obj: Any, v: Visitor, *args, **kwargs):
    return obj._accept(v, *args, **kwargs)



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import numpy as np

import lale.docstrings
import lale.helpers
import lale.operators
from lale.expressions import count, it
from lale.expressions import sum as agg_sum
from lale.lib.sklearn import standard_scaler

from .aggregate import Aggregate
from .map import Map


class _StandardScalerImpl:
    def __init__(self, *, copy=True, with_mean=True, with_std=True):
        self._hyperparams = {"copy": copy, "with_mean": with_mean, "with_std": with_std}
        self.n_samples_seen_ = 0

    def fit(self, X, y=None):
        self._set_fit_attributes(self._lift(X, self._hyperparams))
        return self

    def partial_fit(self, X, y=None):
        if self.n_samples_seen_ == 0:  # first fit
            return self.fit(X)
        lifted_a = self.feature_names_in_, self.n_samples_seen_, self._sum1, self._sum2
        lifted_b = self._lift(X, self._hyperparams)
        self._set_fit_attributes(self._combine(lifted_a, lifted_b))
        return self

    def transform(self, X):
        if self._transformer is None:
            self._transformer = self._build_transformer()
        return self._transformer.transform(X)

    def _set_fit_attributes(self, lifted):
        self.feature_names_in_, self.n_samples_seen_, self._sum1, self._sum2 = lifted
        n = self.n_samples_seen_
        if self._hyperparams["with_std"]:
            # Table 1 of http://www.vldb.org/pvldb/vol8/p702-tangwongsan.pdf
            self.var_ = [
                (self._sum2[i] - self._sum1[i] * self._sum1[i] / n) / n
                for i in range(len(self._sum1))
            ]
            self.scale_ = [
                1.0 if self.var_[i] == 0.0 else np.sqrt(self.var_[i])
                for i in range(len(self._sum1))
            ]
        else:
            self.var_ = None
            self.scale_ = None
        if self._hyperparams["with_mean"]:
            self.mean_ = [self._sum1[i] / n for i in range(len(self._sum1))]
        else:
            self.mean_ = None
        self.n_features_in_ = len(self.feature_names_in_)
        self._transformer = None

    def _build_transformer(self):
        def scale_expr(col_idx, col_name):
            expr = it[col_name]
            if self.mean_ is not None:
                expr = expr - self.mean_[col_idx]
            if self.scale_ is not None:
                expr = expr / self.scale_[col_idx]
            return expr

        result = Map(
            columns={
                col_name: scale_expr(col_idx, col_name)
                for col_idx, col_name in enumerate(self.feature_names_in_)
            }
        )
        return result

    @staticmethod
    def _lift(X, hyperparams):
        feature_names_in = X.columns
        count_op = Aggregate(columns={"count": count(it[feature_names_in[0]])})
        count_data = lale.helpers._ensure_pandas(count_op.transform(X))
        n_samples_seen = count_data.loc[0, "count"]
        if hyperparams["with_mean"] or hyperparams["with_std"]:
            sum1_op = Aggregate(columns={c: agg_sum(it[c]) for c in feature_names_in})
            sum1_data = lale.helpers._ensure_pandas(sum1_op.transform(X))
            sum1 = [sum1_data.loc[0, c] for c in feature_names_in]
        else:
            sum1 = None
        if hyperparams["with_std"]:
            sum2_op = Map(
                columns={c: it[c] * it[c] for c in feature_names_in}
            ) >> Aggregate(columns={c: agg_sum(it[c]) for c in feature_names_in})
            sum2_data = lale.helpers._ensure_pandas(sum2_op.transform(X))
            sum2 = [sum2_data.loc[0, c] for c in feature_names_in]
        else:
            sum2 = None
        return feature_names_in, n_samples_seen, sum1, sum2

    @staticmethod
    def _combine(lifted_a, lifted_b):
        feature_names_in_a, n_samples_seen_a, sum1_a, sum2_a = lifted_a
        feature_names_in_b, n_samples_seen_b, sum1_b, sum2_b = lifted_b
        assert list(feature_names_in_a) == list(feature_names_in_b)
        combined_feat = feature_names_in_a
        combined_n_samples_seen = n_samples_seen_a + n_samples_seen_b
        if sum1_a is None:
            combined_sum1 = None
        else:
            assert sum1_b is not None and len(sum1_a) == len(sum1_b)
            combined_sum1 = [sum1_a[i] + sum1_b[i] for i in range(len(sum1_a))]
        if sum2_a is None:
            combined_sum2 = None
        else:
            assert sum2_b is not None and len(sum2_a) == len(sum2_b)
            combined_sum2 = [sum2_a[i] + sum2_b[i] for i in range(len(sum2_a))]
        return combined_feat, combined_n_samples_seen, combined_sum1, combined_sum2


_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Relational algebra reimplementation of scikit-learn's `StandardScaler`_ transformer that standardizes features by removing the mean and scaling to unit variance.
Works on both pandas and Spark dataframes by using `Aggregate`_ for `fit` and `Map`_ for `transform`, which in turn use the appropriate backend.

.. _`StandardScaler`: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html
.. _`Aggregate`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.aggregate.html
.. _`Map`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.map.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.standard_scaler.html",
    "type": "object",
    "tags": {
        "pre": ["~categoricals"],
        "op": ["transformer", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": standard_scaler._hyperparams_schema,
        "input_fit": standard_scaler._input_fit_schema,
        "input_transform": standard_scaler._input_transform_schema,
        "output_transform": standard_scaler._output_transform_schema,
    },
}

StandardScaler = lale.operators.make_operator(_StandardScalerImpl, _combined_schemas)

StandardScaler = typing.cast(
    lale.operators.PlannedIndividualOp,
    StandardScaler.customize_schema(
        copy={
            "enum": [True],
            "description": "This implementation only supports `copy=True`.",
            "default": True,
        },
    ),
)

lale.docstrings.set_docstrings(StandardScaler)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn
import sklearn.preprocessing

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                "func",
                "inverse_func",
                "validate",
                "accept_sparse",
                "pass_y",
                "check_inverse",
                "kw_args",
                "inv_kw_args",
            ],
            "relevantToOptimizer": [],
            "properties": {
                "func": {
                    "anyOf": [{"laleType": "callable"}, {"enum": [None]}],
                    "default": None,
                    "description": "The callable to use for the transformation.",
                },
                "inverse_func": {
                    "anyOf": [{"laleType": "callable"}, {"enum": [None]}],
                    "default": None,
                    "description": "The callable to use for the inverse transformation.",
                },
                "validate": {
                    "type": "boolean",
                    "default": True,
                    "description": "Indicate that the input X array should be checked before calling ``func``.",
                },
                "accept_sparse": {
                    "type": "boolean",
                    "default": False,
                    "description": "Indicate that func accepts a sparse matrix as input.",
                },
                "pass_y": {
                    "anyOf": [{"type": "boolean"}, {"enum": ["deprecated"]}],
                    "default": "deprecated",
                    "description": "Indicate that transform should forward the y argument to the inner callable.",
                },
                "check_inverse": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to check that ``func`` followed by ``inverse_func`` leads to the original inputs.",
                },
                "kw_args": {
                    "anyOf": [{"type": "object"}, {"enum": [None]}],
                    "default": None,
                    "description": "Dictionary of additional keyword arguments to pass to func.",
                },
                "inv_kw_args": {
                    "anyOf": [{"type": "object"}, {"enum": [None]}],
                    "default": None,
                    "description": "Dictionary of additional keyword arguments to pass to inverse_func.",
                },
            },
        },
        {
            "description": "If validate is False, then accept_sparse has no effect. Otherwise, if accept_sparse is false, sparse matrix inputs will cause an exception to be raised.",
            "anyOf": [
                {"type": "object", "properties": {"validate": {"enum": [False]}}},
                {"type": "object", "laleNot": "X/isSparse"},
                {
                    "type": "object",
                    "properties": {"accept_sparse": {"enum": [True]}},
                },
            ],
        },
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
            },
        },
        "y": {"laleType": "Any"},
    },
}

_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
            },
        }
    },
}

_output_transform_schema = {"type": "array", "items": {"laleType": "Any"}}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """FunctionTransformer_ from scikit-learn constructs a transformer from an arbitrary callable that operates at the level of an entire dataset.

.. _FunctionTransformer: https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.FunctionTransformer.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.sklearn.function_transformer.html",
    "import_from": "sklearn.preprocessing",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}

FunctionTransformer: lale.operators.PlannedIndividualOp
FunctionTransformer = lale.operators.make_operator(
    sklearn.preprocessing.FunctionTransformer, _combined_schemas
)

if sklearn.__version__ >= "0.22":
    # old: https://scikit-learn.org/0.20/modules/generated/sklearn.preprocessing.FunctionTransformer.html
    # new: https://scikit-learn.org/0.23/modules/generated/sklearn.preprocessing.FunctionTransformer.html
    from lale.schemas import Bool

    FunctionTransformer = FunctionTransformer.customize_schema(
        validate=Bool(
            desc="Indicate that the input X array should be checked before calling ``func``.",
            default=False,
        ),
        pass_y=None,
        set_as_available=True,
    )


lale.docstrings.set_docstrings(FunctionTransformer)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from test import EnableSchemaValidation
from test.mock_module import UnknownOp

import numpy as np
from lightgbm import LGBMClassifier as baz
from sklearn.decomposition import PCA as foo
from sklearn.linear_model import Lars as foobar
from xgboost import XGBClassifier as bar

import lale
import lale.schemas as schemas
import lale.type_checking
from lale.search.lale_grid_search_cv import get_grid_search_parameter_grids


class TestCustomSchema(unittest.TestCase):
    def setUp(self):
        import sklearn.decomposition

        import lale.lib.sklearn
        from lale.operators import make_operator

        self.sk_pca = make_operator(sklearn.decomposition.PCA, schemas={})
        self.ll_pca = lale.lib.sklearn.PCA
        self.maxDiff = None

    def test_override_schemas(self):
        with EnableSchemaValidation():
            init_schemas = self.sk_pca._schemas
            pca_schemas = self.ll_pca._schemas
            foo = self.sk_pca.customize_schema(schemas=schemas.JSON(pca_schemas))
            self.assertEqual(foo._schemas, pca_schemas)
            self.assertEqual(self.sk_pca._schemas, init_schemas)
            self.assertRaises(Exception, self.sk_pca.customize_schema, schemas={})

    def test_override_input(self):
        with EnableSchemaValidation():
            init_input_schema = self.sk_pca.get_schema("input_fit")
            pca_input = self.ll_pca.get_schema("input_fit")
            foo = self.sk_pca.customize_schema(input_fit=schemas.JSON(pca_input))
            self.assertEqual(foo.get_schema("input_fit"), pca_input)
            lale.type_checking.validate_is_schema(foo._schemas)
            self.assertEqual(self.sk_pca.get_schema("input_fit"), init_input_schema)
            self.assertRaises(Exception, self.sk_pca.customize_schema, input_fit=42)
            _ = self.sk_pca.customize_schema(input_foo=pca_input)

    def test_override_output(self):
        with EnableSchemaValidation():
            init_output_schema = self.sk_pca.get_schema("output_transform")
            pca_output = self.ll_pca.get_schema("output_transform")
            foo = self.sk_pca.customize_schema(
                output_transform=schemas.JSON(pca_output)
            )
            self.assertEqual(foo.get_schema("output_transform"), pca_output)
            lale.type_checking.validate_is_schema(foo._schemas)
            self.assertEqual(
                self.sk_pca.get_schema("output_transform"), init_output_schema
            )
            self.assertRaises(Exception, self.sk_pca.customize_schema, output=42)
            _ = self.sk_pca.customize_schema(output_foo=pca_output)

    def test_override_output2(self):
        init_output_schema = self.sk_pca.get_schema("output_transform")
        pca_output = schemas.AnyOf(
            [
                schemas.Array(schemas.Array(schemas.Float())),
                schemas.Array(schemas.Float()),
            ]
        )
        expected = {
            "anyOf": [
                {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                },
                {"type": "array", "items": {"type": "number"}},
            ]
        }
        foo = self.sk_pca.customize_schema(output_transform=pca_output)
        self.assertEqual(foo.get_schema("output_transform"), expected)
        lale.type_checking.validate_is_schema(foo._schemas)
        self.assertEqual(self.sk_pca.get_schema("output_transform"), init_output_schema)

    def test_override_bool_param_sk(self):
        with EnableSchemaValidation():
            init = self.sk_pca.hyperparam_schema("whiten")
            expected = {"default": True, "type": "boolean", "description": "override"}
            foo = self.sk_pca.customize_schema(
                whiten=schemas.Bool(default=True, desc="override")
            )
            self.assertEqual(foo.hyperparam_schema("whiten"), expected)
            lale.type_checking.validate_is_schema(foo._schemas)
            self.assertEqual(self.sk_pca.hyperparam_schema("whiten"), init)
            self.assertRaises(Exception, self.sk_pca.customize_schema, whitenX=42)

    def test_override_bool_param_ll(self):
        with EnableSchemaValidation():
            init = self.ll_pca.hyperparam_schema("whiten")
            expected = {"default": True, "type": "boolean"}
            foo = self.ll_pca.customize_schema(whiten=schemas.Bool(default=True))
            self.assertEqual(foo.hyperparam_schema("whiten"), expected)
            lale.type_checking.validate_is_schema(foo._schemas)
            self.assertEqual(self.ll_pca.hyperparam_schema("whiten"), init)
            self.assertRaises(Exception, self.ll_pca.customize_schema, whitenX=42)

    def test_override_enum_param(self):
        with EnableSchemaValidation():
            init = self.ll_pca.hyperparam_schema("svd_solver")
            expected = {"default": "full", "enum": ["auto", "full"]}
            foo = self.ll_pca.customize_schema(
                svd_solver=schemas.Enum(default="full", values=["auto", "full"])
            )
            self.assertEqual(foo.hyperparam_schema("svd_solver"), expected)
            lale.type_checking.validate_is_schema(foo._schemas)
            self.assertEqual(self.ll_pca.hyperparam_schema("svd_solver"), init)

    def test_override_float_param(self):
        init = self.ll_pca.hyperparam_schema("tol")
        expected = {
            "default": 0.1,
            "type": "number",
            "minimum": -10,
            "maximum": 10,
            "exclusiveMaximum": True,
            "exclusiveMinimum": False,
        }
        foo = self.ll_pca.customize_schema(
            tol=schemas.Float(
                default=0.1,
                minimum=-10,
                maximum=10,
                exclusiveMaximum=True,
                exclusiveMinimum=False,
            )
        )
        self.assertEqual(foo.hyperparam_schema("tol"), expected)
        lale.type_checking.validate_is_schema(foo._schemas)
        self.assertEqual(self.ll_pca.hyperparam_schema("tol"), init)

    def test_override_int_param(self):
        init = self.ll_pca.hyperparam_schema("iterated_power")
        expected = {
            "default": 1,
            "type": "integer",
            "minimum": -10,
            "maximum": 10,
            "exclusiveMaximum": True,
            "exclusiveMinimum": False,
        }
        foo = self.ll_pca.customize_schema(
            iterated_power=schemas.Int(
                default=1,
                minimum=-10,
                maximum=10,
                exclusiveMaximum=True,
                exclusiveMinimum=False,
            )
        )
        self.assertEqual(foo.hyperparam_schema("iterated_power"), expected)
        lale.type_checking.validate_is_schema(foo._schemas)
        self.assertEqual(self.ll_pca.hyperparam_schema("iterated_power"), init)

    def test_override_null_param(self):
        init = self.ll_pca.hyperparam_schema("n_components")
        expected = {"enum": [None]}
        foo = self.ll_pca.customize_schema(n_components=schemas.Null())
        self.assertEqual(foo.hyperparam_schema("n_components"), expected)
        lale.type_checking.validate_is_schema(foo._schemas)
        self.assertEqual(self.ll_pca.hyperparam_schema("n_components"), init)

    def test_override_json_param(self):
        init = self.ll_pca.hyperparam_schema("tol")
        expected = {
            "description": "Tol",
            "type": "number",
            "minimum": 0.2,
            "default": 1.0,
        }
        foo = self.ll_pca.customize_schema(tol=schemas.JSON(expected))
        self.assertEqual(foo.hyperparam_schema("tol"), expected)
        lale.type_checking.validate_is_schema(foo._schemas)
        self.assertEqual(self.ll_pca.hyperparam_schema("tol"), init)

    def test_override_any_param(self):
        init = self.ll_pca.hyperparam_schema("iterated_power")
        expected = {
            "anyOf": [{"type": "integer"}, {"enum": ["auto", "full"]}],
            "default": "auto",
        }
        foo = self.ll_pca.customize_schema(
            iterated_power=schemas.AnyOf(
                [schemas.Int(), schemas.Enum(["auto", "full"])], default="auto"
            )
        )
        self.assertEqual(foo.hyperparam_schema("iterated_power"), expected)
        lale.type_checking.validate_is_schema(foo._schemas)
        self.assertEqual(self.ll_pca.hyperparam_schema("iterated_power"), init)

    def test_override_array_param(self):
        init = self.sk_pca.hyperparam_schema("copy")
        expected = {
            "type": "array",
            "minItems": 1,
            "maxItems": 20,
            "items": {"type": "integer"},
        }
        foo = self.sk_pca.customize_schema(
            copy=schemas.Array(minItems=1, maxItems=20, items=schemas.Int())
        )
        self.assertEqual(foo.hyperparam_schema("copy"), expected)
        lale.type_checking.validate_is_schema(foo._schemas)
        self.assertEqual(self.sk_pca.hyperparam_schema("copy"), init)

    def test_override_object_param(self):
        init = self.sk_pca.get_schema("input_fit")
        expected = {
            "type": "object",
            "required": ["X"],
            "additionalProperties": False,
            "properties": {"X": {"type": "array", "items": {"type": "number"}}},
        }
        foo = self.sk_pca.customize_schema(
            input_fit=schemas.Object(
                required=["X"],
                additionalProperties=False,
                X=schemas.Array(schemas.Float()),
            )
        )
        self.assertEqual(foo.get_schema("input_fit"), expected)
        lale.type_checking.validate_is_schema(foo.get_schema("input_fit"))
        self.assertEqual(self.sk_pca.get_schema("input_fit"), init)

    def test_add_constraint(self):
        init = self.sk_pca.hyperparam_schema()
        init_expected = {
            "allOf": [
                {
                    "type": "object",
                    "relevantToOptimizer": [],
                    "additionalProperties": False,
                    "properties": {
                        "n_components": {"default": None},
                        "copy": {"default": True},
                        "whiten": {"default": False},
                        "svd_solver": {"default": "auto"},
                        "tol": {"default": 0.0},
                        "iterated_power": {"default": "auto"},
                        "random_state": {"default": None},
                    },
                }
            ]
        }
        self.assertEqual(init, init_expected)
        expected = {
            "allOf": [
                init_expected["allOf"][0],
                {
                    "anyOf": [
                        {
                            "type": "object",
                            "properties": {
                                "n_components": {
                                    "not": {"enum": ["mle"]},
                                }
                            },
                        },
                        {
                            "type": "object",
                            "properties": {
                                "svd_solver": {"enum": ["full", "auto"]},
                            },
                        },
                    ]
                },
            ]
        }
        foo = self.sk_pca.customize_schema(
            constraint=schemas.AnyOf(
                [
                    schemas.Object(n_components=schemas.Not(schemas.Enum(["mle"]))),
                    schemas.Object(svd_solver=schemas.Enum(["full", "auto"])),
                ]
            )
        )
        self.assertEqual(foo.hyperparam_schema(), expected)
        lale.type_checking.validate_is_schema(foo._schemas)
        self.assertEqual(self.sk_pca.hyperparam_schema(), init)

    def test_add_multiple_constraints(self):
        init = self.sk_pca.hyperparam_schema()
        init_expected = {
            "allOf": [
                {
                    "type": "object",
                    "relevantToOptimizer": [],
                    "additionalProperties": False,
                    "properties": {
                        "n_components": {"default": None},
                        "copy": {"default": True},
                        "whiten": {"default": False},
                        "svd_solver": {"default": "auto"},
                        "tol": {"default": 0.0},
                        "iterated_power": {"default": "auto"},
                        "random_state": {"default": None},
                    },
                }
            ]
        }
        self.assertEqual(init, init_expected)
        expected = {
            "allOf": [
                init_expected["allOf"][0],
                {
                    "anyOf": [
                        {
                            "type": "object",
                            "properties": {
                                "n_components": {
                                    "not": {"enum": ["mle"]},
                                }
                            },
                        },
                        {
                            "type": "object",
                            "properties": {
                                "svd_solver": {"enum": ["full", "auto"]},
                            },
                        },
                    ]
                },
                {
                    "anyOf": [
                        {
                            "type": "object",
                            "properties": {"copy": {"enum": [False]}},
                        },
                        {
                            "type": "object",
                            "properties": {
                                "whiten": {"enum": [False]},
                            },
                        },
                    ]
                },
            ]
        }
        foo = self.sk_pca.customize_schema(
            constraint=[
                schemas.AnyOf(
                    [
                        schemas.Object(n_components=schemas.Not(schemas.Enum(["mle"]))),
                        schemas.Object(svd_solver=schemas.Enum(["full", "auto"])),
                    ]
                ),
                schemas.AnyOf(
                    [
                        schemas.Object(copy=schemas.Enum([False])),
                        schemas.Object(whiten=schemas.Enum([False])),
                    ]
                ),
            ]
        )
        self.assertEqual(foo.hyperparam_schema(), expected)
        lale.type_checking.validate_is_schema(foo._schemas)
        self.assertEqual(self.sk_pca.hyperparam_schema(), init)

    def test_override_relevant(self):
        init = self.ll_pca.hyperparam_schema()["allOf"][0]["relevantToOptimizer"]
        expected = ["svd_solver"]
        foo = self.ll_pca.customize_schema(relevantToOptimizer=["svd_solver"])
        self.assertEqual(
            foo.hyperparam_schema()["allOf"][0]["relevantToOptimizer"], expected
        )
        lale.type_checking.validate_is_schema(foo._schemas)
        self.assertEqual(
            self.ll_pca.hyperparam_schema()["allOf"][0]["relevantToOptimizer"], init
        )
        self.assertRaises(
            Exception, self.sk_pca.customize_schema, relevantToOptimizer={}
        )

    def test_override_tags(self):
        with EnableSchemaValidation():
            init = self.ll_pca._schemas["tags"]
            tags = {
                "pre": ["~categoricals"],
                "op": ["estimator", "classifier", "interpretable"],
                "post": ["probabilities"],
            }
            foo = self.ll_pca.customize_schema(tags=tags)
            self.assertEqual(foo._schemas["tags"], tags)
            lale.type_checking.validate_is_schema(foo._schemas)
            self.assertEqual(self.ll_pca._schemas["tags"], init)
            self.assertRaises(Exception, self.sk_pca.customize_schema, tags=42)

    def test_wrap_imported_operators(self):
        old_globals = {**globals()}
        try:
            from lale.lib.autogen import Lars
            from lale.lib.lightgbm import LGBMClassifier
            from lale.lib.xgboost import XGBClassifier

            lale.wrap_imported_operators(exclude_classes=["foo"])
            from sklearn.decomposition import PCA as sklearn_pca

            op_obj = foo()
            self.assertIsInstance(op_obj, sklearn_pca)
            self.assertEqual(bar._schemas, XGBClassifier._schemas)  # type: ignore
            self.assertEqual(baz._schemas, LGBMClassifier._schemas)  # type: ignore
            self.assertEqual(foobar._schemas, Lars._schemas)
        finally:
            for sym, obj in old_globals.items():
                globals()[sym] = obj


class TestConstraintDropping(unittest.TestCase):
    def test_constraint_dropping(self):
        from lale.lib.sklearn import LogisticRegression
        from lale.operators import make_operator
        from lale.search.schema2search_space import op_to_search_space

        orig_schemas = LogisticRegression._schemas
        mod_schemas = {
            **orig_schemas,
            "properties": {
                **orig_schemas["properties"],
                "hyperparams": {
                    "allOf": [
                        s if i == 0 else {**s, "forOptimizer": False}
                        for i, s in enumerate(
                            orig_schemas["properties"]["hyperparams"]["allOf"]
                        )
                    ]
                },
            },
        }
        orig_space = op_to_search_space(LogisticRegression)
        mod_op = make_operator(LogisticRegression._impl_class(), mod_schemas)
        mod_space = op_to_search_space(mod_op)
        # dropping constraints makes the search space smaller
        self.assertGreater(len(str(orig_space)), len(str(mod_space)))


class TestConstraintMerging(unittest.TestCase):
    def test_override_float_param1(self):
        from lale.lib.sklearn import PCA
        from lale.search.schema2search_space import op_to_search_space
        from lale.search.search_space import SearchSpaceNumber, SearchSpaceObject

        pca = PCA.customize_schema(
            relevantToOptimizer=["tol"],
            tol=schemas.Float(
                minimum=0.25,
                minimumForOptimizer=0,
                maximum=0.5,
                maximumForOptimizer=1.0,
                exclusiveMaximum=True,
                exclusiveMinimum=False,
            ),
        )
        search = op_to_search_space(pca)
        assert isinstance(search, SearchSpaceObject)
        num_space = list(search.choices)[0][0]
        assert isinstance(num_space, SearchSpaceNumber)
        self.assertEqual(num_space.minimum, 0.25)
        self.assertEqual(num_space.maximum, 0.5)
        self.assertTrue(num_space.exclusiveMaximum)
        self.assertFalse(num_space.exclusiveMinimum)

    def test_override_float_param2(self):
        from lale.lib.sklearn import PCA
        from lale.search.schema2search_space import op_to_search_space
        from lale.search.search_space import SearchSpaceNumber, SearchSpaceObject

        pca = PCA.customize_schema(
            relevantToOptimizer=["tol"],
            tol=schemas.Float(
                minimum=0,
                minimumForOptimizer=0.25,
                maximum=1.5,
                maximumForOptimizer=1.0,
                exclusiveMaximumForOptimizer=False,
                exclusiveMinimumForOptimizer=True,
            ),
        )
        search = op_to_search_space(pca)
        assert isinstance(search, SearchSpaceObject)
        num_space = list(search.choices)[0][0]
        assert isinstance(num_space, SearchSpaceNumber)
        self.assertEqual(num_space.minimum, 0.25)
        self.assertEqual(num_space.maximum, 1.0)
        self.assertFalse(num_space.exclusiveMaximum)
        self.assertTrue(num_space.exclusiveMinimum)

    def test_override_int_param1(self):
        from lale.lib.sklearn import PCA
        from lale.search.schema2search_space import op_to_search_space
        from lale.search.search_space import SearchSpaceNumber, SearchSpaceObject

        pca = PCA.customize_schema(
            relevantToOptimizer=["iterated_power"],
            iterated_power=schemas.Float(
                minimum=1,
                minimumForOptimizer=0,
                maximum=5,
                maximumForOptimizer=6,
                exclusiveMaximum=True,
                exclusiveMinimum=False,
            ),
        )
        search = op_to_search_space(pca)
        assert isinstance(search, SearchSpaceObject)
        num_space = list(search.choices)[0][0]
        assert isinstance(num_space, SearchSpaceNumber)
        self.assertEqual(num_space.minimum, 1)
        self.assertEqual(num_space.maximum, 5)
        self.assertTrue(num_space.exclusiveMaximum)
        self.assertFalse(num_space.exclusiveMinimum)

    def test_override_int_param2(self):
        from lale.lib.sklearn import PCA
        from lale.search.schema2search_space import op_to_search_space
        from lale.search.search_space import SearchSpaceNumber, SearchSpaceObject

        pca = PCA.customize_schema(
            relevantToOptimizer=["iterated_power"],
            iterated_power=schemas.Float(
                minimum=0,
                minimumForOptimizer=1,
                maximum=6,
                maximumForOptimizer=5,
                exclusiveMaximumForOptimizer=False,
                exclusiveMinimumForOptimizer=True,
            ),
        )
        search = op_to_search_space(pca)
        assert isinstance(search, SearchSpaceObject)
        num_space = list(search.choices)[0][0]
        assert isinstance(num_space, SearchSpaceNumber)
        self.assertEqual(num_space.minimum, 1)
        self.assertEqual(num_space.maximum, 5)
        self.assertFalse(num_space.exclusiveMaximum)
        self.assertTrue(num_space.exclusiveMinimum)


class TestWrapUnknownOps(unittest.TestCase):
    expected_schema = {
        "allOf": [
            {
                "type": "object",
                "relevantToOptimizer": [],
                "additionalProperties": False,
                "properties": {
                    "n_neighbors": {"default": 5},
                    "algorithm": {"default": "auto"},
                },
            }
        ]
    }

    def test_wrap_from_class(self):
        from lale.operators import PlannedIndividualOp, make_operator

        self.assertFalse(isinstance(UnknownOp, PlannedIndividualOp))
        Wrapped = make_operator(UnknownOp)
        self.assertTrue(isinstance(Wrapped, PlannedIndividualOp))
        self.assertEqual(Wrapped.hyperparam_schema(), self.expected_schema)
        instance = Wrapped(n_neighbors=3)
        self.assertEqual(instance.hyperparams(), {"n_neighbors": 3})

    def test_wrapped_from_import(self):
        old_globals = {**globals()}
        try:
            from lale.operators import PlannedIndividualOp

            self.assertFalse(isinstance(UnknownOp, PlannedIndividualOp))
            lale.wrap_imported_operators()
            self.assertFalse(isinstance(UnknownOp, PlannedIndividualOp))
        finally:
            for sym, obj in old_globals.items():
                globals()[sym] = obj

    def test_wrap_from_instance(self):
        from sklearn.base import clone

        from lale.operators import TrainableIndividualOp, make_operator

        self.assertFalse(isinstance(UnknownOp, TrainableIndividualOp))
        instance = UnknownOp(n_neighbors=3)
        self.assertFalse(isinstance(instance, TrainableIndividualOp))
        wrapped = make_operator(instance)
        self.assertTrue(isinstance(wrapped, TrainableIndividualOp))
        assert isinstance(
            wrapped, TrainableIndividualOp
        )  # help type checkers that don't know about assertTrue
        self.assertEqual(wrapped.reduced_hyperparams(), {"n_neighbors": 3})
        cloned = clone(wrapped)
        self.assertTrue(isinstance(cloned, TrainableIndividualOp))
        self.assertEqual(cloned.reduced_hyperparams(), {"n_neighbors": 3})


class TestFreeze(unittest.TestCase):
    def test_individual_op_freeze_trainable(self):
        from lale.lib.sklearn import LogisticRegression

        liquid = LogisticRegression(C=0.1, solver="liblinear")
        self.assertIn("dual", liquid.free_hyperparams())
        self.assertFalse(liquid.is_frozen_trainable())
        liquid_grid = get_grid_search_parameter_grids(liquid)
        self.assertTrue(len(liquid_grid) > 1, f"grid size {len(liquid_grid)}")
        frozen = liquid.freeze_trainable()
        self.assertEqual(len(frozen.free_hyperparams()), 0)
        self.assertTrue(frozen.is_frozen_trainable())
        frozen_grid = get_grid_search_parameter_grids(frozen)
        self.assertEqual(len(frozen_grid), 1)

    def test_pipeline_freeze_trainable(self):
        from lale.lib.sklearn import PCA, LogisticRegression

        liquid = PCA() >> LogisticRegression()
        self.assertFalse(liquid.is_frozen_trainable())
        liquid_grid = get_grid_search_parameter_grids(liquid)
        self.assertTrue(len(liquid_grid) > 1, f"grid size {len(liquid_grid)}")
        frozen = liquid.freeze_trainable()
        self.assertTrue(frozen.is_frozen_trainable())
        frozen_grid = get_grid_search_parameter_grids(frozen)
        self.assertEqual(len(frozen_grid), 1)

    def test_individual_op_freeze_trained(self):
        from lale.lib.sklearn import KNeighborsClassifier

        with EnableSchemaValidation():
            trainable = KNeighborsClassifier(n_neighbors=1)
            X = np.array([[0.0], [1.0], [2.0]])
            y_old = np.array([0.0, 0.0, 1.0])
            y_new = np.array([1.0, 0.0, 0.0])
            liquid_old = trainable.fit(X, y_old)
            self.assertEqual(list(liquid_old.predict(X)), list(y_old))
            liquid_new = liquid_old.fit(X, y_new)
            self.assertEqual(list(liquid_new.predict(X)), list(y_new))
            frozen_old = trainable.fit(X, y_old).freeze_trained()
            self.assertFalse(liquid_old.is_frozen_trained())
            self.assertTrue(frozen_old.is_frozen_trained())
            self.assertEqual(list(frozen_old.predict(X)), list(y_old))
            frozen_new = frozen_old.fit(X, y_new)
            self.assertEqual(list(frozen_new.predict(X)), list(y_old))

    def test_pipeline_freeze_trained(self):
        from lale.lib.sklearn import LogisticRegression, MinMaxScaler

        trainable = MinMaxScaler() >> LogisticRegression()
        X = [[0.0], [1.0], [2.0]]
        y = [0.0, 0.0, 1.0]
        liquid = trainable.fit(X, y)
        frozen = liquid.freeze_trained()
        self.assertFalse(liquid.is_frozen_trained())
        self.assertTrue(frozen.is_frozen_trained())

    def test_trained_individual_op_freeze_trainable(self):
        from lale.lib.sklearn import KNeighborsClassifier
        from lale.operators import TrainedIndividualOp

        with EnableSchemaValidation():
            trainable = KNeighborsClassifier(n_neighbors=1)
            X = np.array([[0.0], [1.0], [2.0]])
            y_old = np.array([0.0, 0.0, 1.0])
            liquid = trainable.fit(X, y_old)
            self.assertIsInstance(liquid, TrainedIndividualOp)
            self.assertFalse(liquid.is_frozen_trainable())
            self.assertIn("algorithm", liquid.free_hyperparams())
            frozen = liquid.freeze_trainable()
            self.assertIsInstance(frozen, TrainedIndividualOp)
            self.assertTrue(frozen.is_frozen_trainable())
            self.assertFalse(frozen.is_frozen_trained())
            self.assertEqual(len(frozen.free_hyperparams()), 0)

    def test_trained_pipeline_freeze_trainable(self):
        from lale.lib.sklearn import LogisticRegression, MinMaxScaler
        from lale.operators import TrainedPipeline

        trainable = MinMaxScaler() >> LogisticRegression()
        X = [[0.0], [1.0], [2.0]]
        y = [0.0, 0.0, 1.0]
        liquid = trainable.fit(X, y)
        self.assertIsInstance(liquid, TrainedPipeline)
        self.assertFalse(liquid.is_frozen_trainable())
        frozen = liquid.freeze_trainable()
        self.assertFalse(liquid.is_frozen_trainable())
        self.assertTrue(frozen.is_frozen_trainable())
        self.assertIsInstance(frozen, TrainedPipeline)



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast  # see also https://greentreesnakes.readthedocs.io/
import pprint
import typing
from copy import deepcopy
from io import StringIO
from typing import Any, Dict, Optional, Union

import astunparse

AstLits = (ast.Num, ast.Str, ast.List, ast.Tuple, ast.Set, ast.Dict, ast.Constant)
AstLit = Union[ast.Num, ast.Str, ast.List, ast.Tuple, ast.Set, ast.Dict, ast.Constant]
AstExprs = (
    *AstLits,
    ast.Name,
    ast.Expr,
    ast.UnaryOp,
    ast.BinOp,
    ast.BoolOp,
    ast.Compare,
    ast.Call,
    ast.Attribute,
    ast.Subscript,
)
AstExpr = Union[
    AstLit,
    ast.Name,
    ast.Expr,
    ast.UnaryOp,
    ast.BinOp,
    ast.BoolOp,
    ast.Compare,
    ast.Call,
    ast.Attribute,
    ast.Subscript,
]


# !! WORKAROUND !!
# There is a bug with astunparse and Python 3.8.
# https://github.com/simonpercivall/astunparse/issues/43
# Until it is fixed (which may be never), here is a workaround,
# based on the workaround found in https://github.com/juanlao7/codeclose
class FixUnparser(astunparse.Unparser):
    def _Constant(self, t):
        if not hasattr(t, "kind"):
            setattr(t, "kind", None)

        super()._Constant(t)


# !! WORKAROUND !!
# This method should be called instead of astunparse.unparse
def fixedUnparse(tree):
    v = StringIO()
    FixUnparser(tree, file=v)
    return v.getvalue()


class Expr:
    _expr: AstExpr

    def __init__(self, expr: AstExpr, istrue=None):
        # _istrue variable is used to check the boolean nature of
        # '==' and '!=' operator's results.
        self._expr = expr
        self._istrue = istrue

    def __bool__(self) -> bool:
        if self._istrue is not None:
            return self._istrue
        raise TypeError(
            f"Cannot convert expression e1=`{str(self)}` to bool."
            "Instead of `e1 and e2`, try writing `[e1, e2]`."
        )

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result

    def __eq__(self, other):
        if isinstance(other, Expr):
            comp = ast.Compare(
                left=self._expr, ops=[ast.Eq()], comparators=[other._expr]
            )
            return Expr(comp, istrue=self is other)
        elif other is not None:
            comp = ast.Compare(
                left=self._expr, ops=[ast.Eq()], comparators=[ast.Constant(value=other)]
            )
            return Expr(comp, istrue=False)
        else:
            return False

    def __ge__(self, other):
        if isinstance(other, Expr):
            comp = ast.Compare(
                left=self._expr, ops=[ast.GtE()], comparators=[other._expr]
            )
            return Expr(comp)
        elif other is not None:
            comp = ast.Compare(
                left=self._expr,
                ops=[ast.GtE()],
                comparators=[ast.Constant(value=other)],
            )
            return Expr(comp)
        else:
            return False

    def __getattr__(self, name: str) -> "Expr":
        attr = ast.Attribute(value=self._expr, attr=name)
        return Expr(attr)

    def __getitem__(self, key: Union[int, str, slice]) -> "Expr":
        key_ast: Union[ast.Index, ast.Slice]
        if isinstance(key, int):
            key_ast = ast.Index(ast.Num(n=key))
        elif isinstance(key, str):
            key_ast = ast.Index(ast.Str(s=key))
        elif isinstance(key, slice):
            key_ast = ast.Slice(key.start, key.stop, key.step)
        else:
            raise TypeError(f"expected int, str, or slice, got {type(key)}")
        subscript = ast.Subscript(value=self._expr, slice=key_ast)
        return Expr(subscript)

    def __gt__(self, other):
        if isinstance(other, Expr):
            comp = ast.Compare(
                left=self._expr, ops=[ast.Gt()], comparators=[other._expr]
            )
            return Expr(comp)
        elif other is not None:
            comp = ast.Compare(
                left=self._expr, ops=[ast.Gt()], comparators=[ast.Constant(value=other)]
            )
            return Expr(comp)
        else:
            return False

    def __le__(self, other):
        if isinstance(other, Expr):
            comp = ast.Compare(
                left=self._expr, ops=[ast.LtE()], comparators=[other._expr]
            )
            return Expr(comp)
        elif other is not None:
            comp = ast.Compare(
                left=self._expr,
                ops=[ast.LtE()],
                comparators=[ast.Constant(value=other)],
            )
            return Expr(comp)
        else:
            return False

    def __lt__(self, other):
        if isinstance(other, Expr):
            comp = ast.Compare(
                left=self._expr, ops=[ast.Lt()], comparators=[other._expr]
            )
            return Expr(comp)
        elif other is not None:
            comp = ast.Compare(
                left=self._expr, ops=[ast.Lt()], comparators=[ast.Constant(value=other)]
            )
            return Expr(comp)
        else:
            return False

    def __ne__(self, other):
        if isinstance(other, Expr):
            comp = ast.Compare(
                left=self._expr, ops=[ast.NotEq()], comparators=[other._expr]
            )
            return Expr(comp, istrue=self is other)
        elif other is not None:
            comp = ast.Compare(
                left=self._expr,
                ops=[ast.NotEq()],
                comparators=[ast.Constant(value=other)],
            )
            return Expr(comp, istrue=False)
        else:
            return False

    def __str__(self) -> str:
        result = fixedUnparse(self._expr).strip()
        if isinstance(self._expr, (ast.UnaryOp, ast.BinOp, ast.Compare, ast.BoolOp)):
            if result.startswith("(") and result.endswith(")"):
                result = result[1:-1]
        return result

    def __add__(self, other):
        return _make_binop(ast.Add(), self._expr, other)

    def __sub__(self, other):
        return _make_binop(ast.Sub(), self._expr, other)

    def __mul__(self, other):
        return _make_binop(ast.Mult(), self._expr, other)

    def __truediv__(self, other):
        return _make_binop(ast.Div(), self._expr, other)

    def __floordiv__(self, other):
        return _make_binop(ast.FloorDiv(), self._expr, other)

    def __mod__(self, other):
        return _make_binop(ast.Mod(), self._expr, other)

    def __pow__(self, other):
        return _make_binop(ast.Pow(), self._expr, other)


def _make_binop(op, left, other):
    if isinstance(other, Expr):
        e = ast.BinOp(left=left, op=op, right=other._expr)
        return Expr(e)
    elif other is not None:
        e = ast.BinOp(left=left, op=op, right=ast.Constant(value=other))
        return Expr(e)
    else:
        return False


def _make_ast_expr(arg: Union[None, Expr, int, float, str, AstExpr]) -> AstExpr:
    if arg is None:
        return ast.Constant(value=None)
    elif isinstance(arg, Expr):
        return arg._expr
    elif isinstance(arg, (int, float)):
        return ast.Num(n=arg)
    elif isinstance(arg, str):
        return ast.Str(s=arg)
    else:
        assert isinstance(arg, AstExprs), type(arg)
        return arg


def _make_call_expr(name: str, *args: Union[Expr, AstExpr, int, str]) -> Expr:
    func_ast = ast.Name(id=name)
    args_asts = [_make_ast_expr(arg) for arg in args]
    call_ast = ast.Call(func=func_ast, args=args_asts, keywords=[])
    return Expr(call_ast)


def string_indexer(subject: Expr) -> Expr:
    return _make_call_expr("string_indexer", subject)


def collect_set(group: Expr) -> Expr:
    return _make_call_expr("collect_set", group)


def count(group: Expr) -> Expr:
    return _make_call_expr("count", group)


def day_of_month(subject: Expr, fmt: Optional[str] = None) -> Expr:
    if fmt is None:
        return _make_call_expr("day_of_month", subject)
    return _make_call_expr("day_of_month", subject, fmt)


def day_of_week(subject: Expr, fmt: Optional[str] = None) -> Expr:
    if fmt is None:
        return _make_call_expr("day_of_week", subject)
    return _make_call_expr("day_of_week", subject, fmt)


def day_of_year(subject: Expr, fmt: Optional[str] = None) -> Expr:
    if fmt is None:
        return _make_call_expr("day_of_year", subject)
    return _make_call_expr("day_of_year", subject, fmt)


def distinct_count(group: Expr) -> Expr:
    return _make_call_expr("distinct_count", group)


def hour(subject: Expr, fmt: Optional[str] = None) -> Expr:
    if fmt is None:
        return _make_call_expr("hour", subject)
    return _make_call_expr("hour", subject, fmt)


def item(group: Expr, value: Union[int, str]) -> Expr:
    return _make_call_expr("item", group, value)


def max(group: Expr) -> Expr:
    return _make_call_expr("max", group)


def max_gap_to_cutoff(group: Expr, cutoff: Expr) -> Expr:
    return _make_call_expr("max_gap_to_cutoff", group, cutoff)


def mean(group: Expr) -> Expr:
    return _make_call_expr("mean", group)


def min(group: Expr) -> Expr:
    return _make_call_expr("min", group)


def minute(subject: Expr, fmt: Optional[str] = None) -> Expr:
    if fmt is None:
        return _make_call_expr("minute", subject)
    return _make_call_expr("minute", subject, fmt)


def month(subject: Expr, fmt: Optional[str] = None) -> Expr:
    if fmt is None:
        return _make_call_expr("month", subject)
    return _make_call_expr("month", subject, fmt)


def normalized_count(group: Expr) -> Expr:
    return _make_call_expr("normalized_count", group)


def normalized_sum(group: Expr) -> Expr:
    return _make_call_expr("normalized_sum", group)


def recent(series: Expr, age: int) -> Expr:
    return _make_call_expr("recent", series, age)


def recent_gap_to_cutoff(series: Expr, cutoff: Expr, age: int) -> Expr:
    return _make_call_expr("recent_gap_to_cutoff", series, cutoff, age)


def replace(
    subject: Expr,
    old2new: Dict[Any, Any],
    handle_unknown="identity",
    unknown_value=None,
) -> Expr:
    old2new_str = pprint.pformat(old2new)
    module_ast = ast.parse(old2new_str)
    old2new_ast = typing.cast(ast.Expr, module_ast.body[0])
    assert handle_unknown in ["identity", "use_encoded_value"]
    return _make_call_expr(
        "replace",
        subject,
        old2new_ast,
        handle_unknown,
        unknown_value,
    )


def identity(subject: Expr) -> Expr:
    return _make_call_expr("identity", subject)


def sum(group: Expr) -> Expr:
    return _make_call_expr("sum", group)


def trend(series: Expr) -> Expr:
    return _make_call_expr("trend", series)


def variance(group: Expr) -> Expr:
    return _make_call_expr("variance", group)


def window_max(series: Expr, size: int) -> Expr:
    return _make_call_expr("window_max", series, size)


def window_max_trend(series: Expr, size: int) -> Expr:
    return _make_call_expr("window_max_trend", series, size)


def window_mean(series: Expr, size: int) -> Expr:
    return _make_call_expr("window_mean", series, size)


def window_mean_trend(series: Expr, size: int) -> Expr:
    return _make_call_expr("window_mean_trend", series, size)


def window_min(series: Expr, size: int) -> Expr:
    return _make_call_expr("window_min", series, size)


def window_min_trend(series: Expr, size: int) -> Expr:
    return _make_call_expr("window_min_trend", series, size)


def window_variance(series: Expr, size: int) -> Expr:
    return _make_call_expr("window_variance", series, size)


def window_variance_trend(series: Expr, size: int) -> Expr:
    return _make_call_expr("window_variance_trend", series, size)


def first(group: Expr) -> Expr:
    return _make_call_expr("first", group)


def isnan(column: Expr) -> Expr:
    return _make_call_expr("isnan", column)


def isnotnan(column: Expr) -> Expr:
    return _make_call_expr("isnotnan", column)


def isnull(column: Expr) -> Expr:
    return _make_call_expr("isnull", column)


def isnotnull(column: Expr) -> Expr:
    return _make_call_expr("isnotnull", column)


def asc(column: Union[Expr, str]) -> Expr:
    return _make_call_expr("asc", column)


def desc(column: Union[Expr, str]) -> Expr:
    return _make_call_expr("desc", column)


def median(group: Expr) -> Expr:
    return _make_call_expr("median", group)


def mode(group: Expr) -> Expr:
    return _make_call_expr("mode", group)


it = Expr(ast.Name(id="it"))


def _it_column(expr):
    if isinstance(expr, ast.Attribute):
        if _is_ast_name_it(expr.value):
            return expr.attr
        else:
            raise ValueError(
                f"Illegal {fixedUnparse(expr)}. Only the access to `it` is supported"
            )
    elif isinstance(expr, ast.Subscript):
        if isinstance(expr.slice, ast.Constant) or (
            _is_ast_name_it(expr.value) and isinstance(expr.slice, ast.Index)
        ):
            if isinstance(expr.slice, ast.Constant):
                return expr.slice.value
            elif isinstance(expr.slice.value, ast.Constant):
                return expr.slice.value.value
            elif isinstance(expr.slice.value, ast.Str):
                return expr.slice.value.s
        else:
            raise ValueError(
                f"Illegal {fixedUnparse(expr)}. Only the access to `it` is supported"
            )
    else:
        raise ValueError(
            f"Illegal {fixedUnparse(expr)}. Only the access to `it` is supported"
        )


def _is_ast_name_it(expr):
    return isinstance(expr, ast.Name) and expr.id == "it"



from numpy import inf, nan
from sklearn.svm import SVR as Op

from lale.docstrings import set_docstrings
from lale.operators import make_operator


class _SVRImpl:
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams
        self._wrapped_model = Op(**self._hyperparams)

    def fit(self, X, y=None):
        if y is not None:
            self._wrapped_model.fit(X, y)
        else:
            self._wrapped_model.fit(X)
        return self

    def predict(self, X):
        return self._wrapped_model.predict(X)


_hyperparams_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "inherited docstring for SVR    Epsilon-Support Vector Regression.",
    "allOf": [
        {
            "type": "object",
            "required": [
                "kernel",
                "degree",
                "gamma",
                "coef0",
                "tol",
                "C",
                "epsilon",
                "shrinking",
                "cache_size",
                "verbose",
                "max_iter",
            ],
            "relevantToOptimizer": [
                "kernel",
                "degree",
                "gamma",
                "tol",
                "epsilon",
                "shrinking",
                "cache_size",
                "max_iter",
            ],
            "additionalProperties": False,
            "properties": {
                "kernel": {
                    "enum": ["linear", "poly", "precomputed", "sigmoid", "rbf"],
                    "default": "rbf",
                    "description": "Specifies the kernel type to be used in the algorithm",
                },
                "degree": {
                    "type": "integer",
                    "minimumForOptimizer": 2,
                    "maximumForOptimizer": 3,
                    "distribution": "uniform",
                    "default": 3,
                    "description": "Degree of the polynomial kernel function ('poly')",
                },
                "gamma": {
                    "anyOf": [
                        {"type": "number", "forOptimizer": False},
                        {"enum": ["scale", "auto"]},
                    ],
                    "default": "scale",
                    "description": "Kernel coefficient for 'rbf', 'poly' and 'sigmoid'",
                },
                "coef0": {
                    "type": "number",
                    "default": 0.0,
                    "description": "Independent term in kernel function",
                },
                "tol": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 0.01,
                    "distribution": "loguniform",
                    "default": 0.001,
                    "description": "Tolerance for stopping criterion.",
                },
                "C": {
                    "type": "number",
                    "default": 1.0,
                    "description": "Penalty parameter C of the error term.",
                },
                "epsilon": {
                    "type": "number",
                    "minimumForOptimizer": 1e-08,
                    "maximumForOptimizer": 1.35,
                    "distribution": "loguniform",
                    "default": 0.1,
                    "description": "Epsilon in the epsilon-SVR model",
                },
                "shrinking": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to use the shrinking heuristic.",
                },
                "cache_size": {
                    "type": "number",
                    "minimumForOptimizer": 0.0,
                    "maximumForOptimizer": 1.0,
                    "distribution": "uniform",
                    "default": 200,
                    "description": "Specify the size of the kernel cache (in MB).",
                },
                "verbose": {
                    "type": "boolean",
                    "default": False,
                    "description": "Enable verbose output",
                },
                "max_iter": {
                    "XXX TODO XXX": "int, optional (default=-1)",
                    "description": "Hard limit on iterations within solver, or -1 for no limit.",
                    "type": "integer",
                    "minimumForOptimizer": 10,
                    "maximumForOptimizer": 1000,
                    "distribution": "uniform",
                    "default": (-1),
                },
            },
        },
        {"XXX TODO XXX": "Parameter: coef0 > only significant in 'poly' and 'sigmoid'"},
    ],
}
_input_fit_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Fit the SVM model according to the given training data.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": "Training vectors, where n_samples is the number of samples and n_features is the number of features",
        },
        "y": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Target values (class labels in classification, real numbers in regression)",
        },
        "sample_weight": {
            "type": "array",
            "items": {"type": "number"},
            "description": "Per-sample weights",
        },
    },
}
_input_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Perform regression on samples in X.",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
            "description": 'For kernel="precomputed", the expected shape of X is (n_samples_test, n_samples_train).',
        }
    },
}
_output_predict_schema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Perform regression on samples in X.",
    "type": "array",
    "items": {"type": "number"},
}
_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "documentation_url": "https://scikit-learn.org/0.20/modules/generated/sklearn.svm.SVR#sklearn-svm-svr",
    "import_from": "sklearn.svm",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}
SVR = make_operator(_SVRImpl, _combined_schemas)

set_docstrings(SVR)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from test import EnableSchemaValidation
from typing import Any

import jsonschema

import lale.lib.lale
import lale.lib.sklearn
import lale.type_checking
from lale.lib.lale import ConcatFeatures
from lale.lib.sklearn import (
    NMF,
    PCA,
    RFE,
    FunctionTransformer,
    LogisticRegression,
    MissingIndicator,
    Nystroem,
    TfidfVectorizer,
)


class TestFeaturePreprocessing(unittest.TestCase):
    def setUp(self):
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split

        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(X, y)


def create_function_test_feature_preprocessor(fproc_name):
    def test_feature_preprocessor(self):
        X_train, y_train = self.X_train, self.y_train
        import importlib

        module_name = ".".join(fproc_name.split(".")[0:-1])
        class_name = fproc_name.split(".")[-1]
        module = importlib.import_module(module_name)

        class_ = getattr(module, class_name)
        fproc = class_()

        from lale.lib.sklearn.one_hot_encoder import OneHotEncoder

        if isinstance(fproc, OneHotEncoder):  # type: ignore
            # fproc = OneHotEncoder(handle_unknown = 'ignore')
            # remove the hack when this is fixed
            fproc = PCA()
        # test_schemas_are_schemas
        lale.type_checking.validate_is_schema(fproc.input_schema_fit())
        lale.type_checking.validate_is_schema(fproc.input_schema_transform())
        lale.type_checking.validate_is_schema(fproc.output_schema_transform())
        lale.type_checking.validate_is_schema(fproc.hyperparam_schema())

        # test_init_fit_transform
        trained = fproc.fit(self.X_train, self.y_train)
        _ = trained.transform(self.X_test)

        # test_predict_on_trainable
        trained = fproc.fit(X_train, y_train)
        fproc.transform(X_train)

        # test_to_json
        fproc.to_json()

        # test_in_a_pipeline
        # This test assumes that the output of feature processing is compatible with LogisticRegression
        from lale.lib.sklearn import LogisticRegression

        pipeline = fproc >> LogisticRegression()
        trained = pipeline.fit(self.X_train, self.y_train)
        _ = trained.predict(self.X_test)

        # Tune the pipeline with LR using Hyperopt
        from lale.lib.lale import Hyperopt

        hyperopt = Hyperopt(estimator=pipeline, max_evals=1, verbose=True, cv=3)
        trained = hyperopt.fit(self.X_train, self.y_train)
        _ = trained.predict(self.X_test)

    test_feature_preprocessor.__name__ = "test_{0}".format(fproc_name.split(".")[-1])
    return test_feature_preprocessor


feature_preprocessors = [
    "lale.lib.sklearn.PolynomialFeatures",
    "lale.lib.sklearn.PCA",
    "lale.lib.sklearn.Nystroem",
    "lale.lib.sklearn.Normalizer",
    "lale.lib.sklearn.MinMaxScaler",
    "lale.lib.sklearn.OneHotEncoder",
    "lale.lib.sklearn.SimpleImputer",
    "lale.lib.sklearn.StandardScaler",
    "lale.lib.sklearn.FeatureAgglomeration",
    "lale.lib.sklearn.RobustScaler",
    "lale.lib.sklearn.QuantileTransformer",
    "lale.lib.sklearn.VarianceThreshold",
    "lale.lib.sklearn.Isomap",
]
for fproc in feature_preprocessors:
    setattr(
        TestFeaturePreprocessing,
        "test_{0}".format(fproc.split(".")[-1]),
        create_function_test_feature_preprocessor(fproc),
    )


class TestNMF(unittest.TestCase):
    def test_init_fit_predict(self):
        import lale.datasets

        nmf = NMF()
        lr = LogisticRegression()
        trainable = nmf >> lr
        (train_X, train_y), (test_X, test_y) = lale.datasets.digits_df()
        trained = trainable.fit(train_X, train_y)
        _ = trained.predict(test_X)

    def test_not_randome_state(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                _ = NMF(random_state='"not RandomState"')


class TestFunctionTransformer(unittest.TestCase):
    def test_init_fit_predict(self):
        import numpy as np

        import lale.datasets

        ft = FunctionTransformer(func=np.log1p)
        lr = LogisticRegression()
        trainable = ft >> lr
        (train_X, train_y), (test_X, test_y) = lale.datasets.digits_df()
        trained = trainable.fit(train_X, train_y)
        _ = trained.predict(test_X)

    def test_not_callable(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                _ = FunctionTransformer(func='"not callable"')


class TestMissingIndicator(unittest.TestCase):
    def test_init_fit_transform(self):
        import numpy as np

        X1 = np.array([[np.nan, 1, 3], [4, 0, np.nan], [8, 1, 0]])
        X2 = np.array([[5, 1, np.nan], [np.nan, 2, 3], [2, 4, 0]])
        trainable = MissingIndicator()
        trained = trainable.fit(X1)
        transformed = trained.transform(X2)
        expected = np.array([[False, True], [True, False], [False, False]])
        self.assertTrue((transformed == expected).all())


class TestRFE(unittest.TestCase):
    def test_init_fit_predict(self):
        import sklearn.datasets
        import sklearn.svm

        svm = lale.lib.sklearn.SVR(kernel="linear")
        rfe = RFE(estimator=svm, n_features_to_select=2)
        lr = LogisticRegression()
        trainable = rfe >> lr
        data = sklearn.datasets.load_iris()
        X, y = data.data, data.target
        trained = trainable.fit(X, y)
        _ = trained.predict(X)

    def test_init_fit_predict_sklearn(self):
        import sklearn.datasets
        import sklearn.svm

        svm = sklearn.svm.SVR(kernel="linear")
        rfe = RFE(estimator=svm, n_features_to_select=2)
        lr = LogisticRegression()
        trainable = rfe >> lr
        data = sklearn.datasets.load_iris()
        X, y = data.data, data.target
        trained = trainable.fit(X, y)
        _ = trained.predict(X)

    def test_not_operator(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                _ = RFE(estimator='"not an operator"', n_features_to_select=2)

    def test_attrib_sklearn(self):
        import sklearn.datasets
        import sklearn.svm

        from lale.lib.sklearn import RFE, LogisticRegression

        svm = sklearn.svm.SVR(kernel="linear")
        rfe = RFE(estimator=svm, n_features_to_select=2)
        lr = LogisticRegression()
        trainable = rfe >> lr
        data = sklearn.datasets.load_iris()
        X, y = data.data, data.target
        trained = trainable.fit(X, y)
        _ = trained.predict(X)
        from lale.lib.lale import Hyperopt

        opt = Hyperopt(estimator=trainable, max_evals=2, verbose=True)
        opt.fit(X, y)

    def test_attrib(self):
        import sklearn.datasets

        from lale.lib.sklearn import RFE, LogisticRegression

        svm = lale.lib.sklearn.SVR(kernel="linear")
        rfe = RFE(estimator=svm, n_features_to_select=2)
        lr = LogisticRegression()
        trainable = rfe >> lr
        data = sklearn.datasets.load_iris()
        X, y = data.data, data.target
        trained = trainable.fit(X, y)
        _ = trained.predict(X)
        from lale.lib.lale import Hyperopt

        opt = Hyperopt(estimator=trainable, max_evals=2, verbose=True)
        opt.fit(X, y)


class TestOrdinalEncoder(unittest.TestCase):
    def setUp(self):
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split

        data = load_iris()
        X, y = data.data, data.target
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(X, y)

    def test_with_hyperopt(self):
        from lale.lib.sklearn import OrdinalEncoder

        fproc = OrdinalEncoder(handle_unknown="ignore")
        from lale.lib.sklearn import LogisticRegression

        pipeline = fproc >> LogisticRegression()

        # Tune the pipeline with LR using Hyperopt
        from lale.lib.lale import Hyperopt

        hyperopt = Hyperopt(estimator=pipeline, max_evals=1)
        trained = hyperopt.fit(self.X_train, self.y_train)
        _ = trained.predict(self.X_test)

    def test_inverse_transform(self):
        from lale.lib.sklearn import OneHotEncoder, OrdinalEncoder

        fproc_ohe = OneHotEncoder(handle_unknown="ignore")
        # test_init_fit_transform
        trained_ohe = fproc_ohe.fit(self.X_train, self.y_train)
        transformed_X = trained_ohe.transform(self.X_test)
        orig_X_ohe = trained_ohe._impl._wrapped_model.inverse_transform(transformed_X)

        fproc_oe = OrdinalEncoder(handle_unknown="ignore")
        # test_init_fit_transform
        trained_oe = fproc_oe.fit(self.X_train, self.y_train)
        transformed_X = trained_oe.transform(self.X_test)
        orig_X_oe = trained_oe._impl.inverse_transform(transformed_X)
        self.assertEqual(orig_X_ohe.all(), orig_X_oe.all())

    def test_handle_unknown_error(self):
        from lale.lib.sklearn import OrdinalEncoder

        fproc_oe = OrdinalEncoder(handle_unknown="error")
        # test_init_fit_transform
        trained_oe = fproc_oe.fit(self.X_train, self.y_train)
        with self.assertRaises(
            ValueError
        ):  # This is repying on the train_test_split, so may fail randomly
            _ = trained_oe.transform(self.X_test)

    def test_encode_unknown_with(self):
        from lale.lib.sklearn import OrdinalEncoder

        fproc_oe = OrdinalEncoder(handle_unknown="ignore", encode_unknown_with=1000)
        # test_init_fit_transform
        trained_oe = fproc_oe.fit(self.X_train, self.y_train)
        transformed_X = trained_oe.transform(self.X_test)
        # This is repying on the train_test_split, so may fail randomly
        self.assertTrue(1000 in transformed_X)
        # Testing that inverse_transform works even for encode_unknown_with=1000
        _ = trained_oe._impl.inverse_transform(transformed_X)


class TestConcatFeatures(unittest.TestCase):
    def test_hyperparam_defaults(self):
        _ = ConcatFeatures()

    def test_init_fit_predict(self):
        trainable_cf = ConcatFeatures()
        A = [[11, 12, 13], [21, 22, 23], [31, 32, 33]]
        B = [[14, 15], [24, 25], [34, 35]]

        trained_cf = trainable_cf.fit(X=[A, B])
        transformed: Any = trained_cf.transform([A, B])
        expected = [[11, 12, 13, 14, 15], [21, 22, 23, 24, 25], [31, 32, 33, 34, 35]]
        for i_sample in range(len(transformed)):
            for i_feature in range(len(transformed[i_sample])):
                self.assertEqual(
                    transformed[i_sample][i_feature], expected[i_sample][i_feature]
                )

    def test_comparison_with_scikit(self):
        import warnings

        warnings.filterwarnings("ignore")
        import sklearn.datasets
        import sklearn.utils

        from lale.helpers import cross_val_score
        from lale.lib.sklearn import PCA

        pca = PCA(n_components=3, random_state=42, svd_solver="arpack")
        nys = Nystroem(n_components=10, random_state=42)
        concat = ConcatFeatures()
        lr = LogisticRegression(random_state=42, C=0.1)
        trainable = (pca & nys) >> concat >> lr
        digits = sklearn.datasets.load_digits()
        X, y = sklearn.utils.shuffle(digits.data, digits.target, random_state=42)

        cv_results = cross_val_score(trainable, X, y)
        cv_results = ["{0:.1%}".format(score) for score in cv_results]

        from sklearn.decomposition import PCA as SklearnPCA
        from sklearn.kernel_approximation import Nystroem as SklearnNystroem
        from sklearn.linear_model import LogisticRegression as SklearnLR
        from sklearn.model_selection import cross_val_score
        from sklearn.pipeline import FeatureUnion, make_pipeline

        union = FeatureUnion(
            [
                (
                    "pca",
                    SklearnPCA(n_components=3, random_state=42, svd_solver="arpack"),
                ),
                ("nys", SklearnNystroem(n_components=10, random_state=42)),
            ]
        )
        lr = SklearnLR(random_state=42, C=0.1)
        pipeline = make_pipeline(union, lr)

        scikit_cv_results = cross_val_score(pipeline, X, y, cv=5)
        scikit_cv_results = ["{0:.1%}".format(score) for score in scikit_cv_results]
        self.assertEqual(cv_results, scikit_cv_results)
        warnings.resetwarnings()

    def test_with_pandas(self):
        import warnings

        from lale.datasets import load_iris_df

        warnings.filterwarnings("ignore")
        pca = PCA(n_components=3)
        nys = Nystroem(n_components=10)
        concat = ConcatFeatures()
        lr = LogisticRegression(random_state=42, C=0.1)
        trainable = (pca & nys) >> concat >> lr

        (X_train, y_train), (X_test, y_test) = load_iris_df()
        trained = trainable.fit(X_train, y_train)
        _ = trained.predict(X_test)

    def test_concat_with_hyperopt(self):
        from lale.lib.lale import Hyperopt

        pca = PCA(n_components=3)
        nys = Nystroem(n_components=10)
        concat = ConcatFeatures()
        lr = LogisticRegression(random_state=42, C=0.1)

        trainable = (pca & nys) >> concat >> lr
        clf = Hyperopt(estimator=trainable, max_evals=2)
        from sklearn.datasets import load_iris

        iris_data = load_iris()
        clf.fit(iris_data.data, iris_data.target)
        clf.predict(iris_data.data)

    def test_concat_with_hyperopt2(self):
        from lale.lib.lale import Hyperopt
        from lale.operators import make_pipeline, make_union

        pca = PCA(n_components=3)
        nys = Nystroem(n_components=10)
        lr = LogisticRegression(random_state=42, C=0.1)

        trainable = make_pipeline(make_union(pca, nys), lr)
        clf = Hyperopt(estimator=trainable, max_evals=2)
        from sklearn.datasets import load_iris

        iris_data = load_iris()
        clf.fit(iris_data.data, iris_data.target)
        clf.predict(iris_data.data)


class TestTfidfVectorizer(unittest.TestCase):
    def test_more_hyperparam_values(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                _ = TfidfVectorizer(
                    max_df=2.5, min_df=2, max_features=1000, stop_words="english"
                )
            with self.assertRaises(jsonschema.ValidationError):
                _ = TfidfVectorizer(
                    max_df=2,
                    min_df=2,
                    max_features=1000,
                    stop_words=["I", "we", "not", "this", "that"],
                    analyzer="char",
                )

    def test_non_null_tokenizer(self):
        # tokenize the doc and lemmatize its tokens
        def my_tokenizer():
            return "abc"

        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                _ = TfidfVectorizer(
                    max_df=2,
                    min_df=2,
                    max_features=1000,
                    stop_words="english",
                    tokenizer=my_tokenizer,
                    analyzer="char",
                )



# Copyright 2020, 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

import lale.docstrings
import lale.operators

from .util import (
    _categorical_fairness_properties,
    _categorical_input_transform_schema,
    _categorical_output_transform_schema,
    _categorical_unsupervised_input_fit_schema,
)


def _redaction_value(column_values):
    unique_values, unique_counts = np.unique(column_values, return_counts=True)
    most_frequent_index = np.argmax(unique_counts)
    most_frequent_value = unique_values[most_frequent_index]
    return most_frequent_value


class _RedactingImpl:
    def __init__(
        self, *, favorable_labels, protected_attributes, unfavorable_labels=None
    ):
        self.prot_attr_names = [pa["feature"] for pa in protected_attributes]

    def fit(self, X, y=None):
        if isinstance(X, pd.DataFrame):
            self.redaction_values = {
                pa: _redaction_value(X[pa]) for pa in self.prot_attr_names
            }
        elif isinstance(X, np.ndarray):
            self.redaction_values = {
                pa: _redaction_value(X[:, pa]) for pa in self.prot_attr_names
            }
        else:
            raise TypeError(f"unexpected type {type(X)}")
        return self

    def transform(self, X):
        if isinstance(X, pd.DataFrame):
            new_columns = [
                (
                    X[name].map(lambda val: self.redaction_values[name])
                    if name in self.redaction_values
                    else X[name]
                )
                for name in X.columns
            ]
            result = pd.concat(new_columns, axis=1)
        elif isinstance(X, np.ndarray):
            result = X.copy()
            for column, value in self.redaction_values.items():
                result[:, column].fill(value)
        else:
            raise TypeError(f"unexpected type {type(X)}")
        return result

    def transform_schema(self, s_X):
        """Used internally by Lale for type-checking downstream operators."""
        return s_X


_input_fit_schema = _categorical_unsupervised_input_fit_schema
_input_transform_schema = _categorical_input_transform_schema
_output_transform_schema = _categorical_output_transform_schema

_hyperparams_schema = {
    "description": "Hyperparameter schema.",
    "allOf": [
        {
            "type": "object",
            "additionalProperties": False,
            "required": [
                "favorable_labels",
                "protected_attributes",
                "unfavorable_labels",
            ],
            "relevantToOptimizer": [],
            "properties": {
                "favorable_labels": {
                    "description": "Ignored.",
                    "laleType": "Any",
                },
                "protected_attributes": _categorical_fairness_properties[
                    "protected_attributes"
                ],
                "unfavorable_labels": {
                    "description": "Ignored.",
                    "laleType": "Any",
                },
            },
        }
    ],
}

_combined_schemas = {
    "description": """Redacting preprocessor for fairness mitigation.

This sets all the protected attributes to constants,
using the most frequent value in the column.
This operator is used internally by various lale.lib.aif360 metrics
and mitigators, so you often do not need to use it directly yourself.
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.aif360.redacting.html#lale.lib.aif360.redacting.Redacting",
    "import_from": "lale.lib.aif360",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


Redacting = lale.operators.make_operator(_RedactingImpl, _combined_schemas)

lale.docstrings.set_docstrings(Redacting)
